﻿using EnvDTE;
using EnvDTE80;
using Microsoft;
using Microsoft.VisualStudio.Shell;
using Microsoft.VisualStudio.Shell.Interop;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace StructLayout
{
    [AttributeUsage(AttributeTargets.Property)]
    public class UIDescription : Attribute
    {
        public string Label { get; set; }
        public string Tooltip { get; set; }
    }

    public class SolutionSettings
    {
        //Parser Settings
        [UIDescription(Label = "Extra Preprocessor Defintions", Tooltip = "Additional preprocessor definitions on top of the auto extracted form the project configuration. (This fields allows $(SolutionDir) style macros)")]
        public string AdditionalPreprocessorDefinitions { set; get; } = "";

        [UIDescription(Label = "Extra Include Dirs", Tooltip = "Additional include directories on top of the auto extracted form the project configuration. (This fields allows $(SolutionDir) style macros)")]
        public string AdditionalIncludeDirs { set; get; } = "";

        [UIDescription(Label = "Extra Force Includes", Tooltip = "Additional files to force include on top of the auto extracted form the project configuration. (This fields allows $(SolutionDir) style macros)")]
        public string AdditionalForceIncludes { set; get; } = "";

        [UIDescription(Label = "Extra Parser Args", Tooltip = "Additional command line arguments passed in to the clang parser. (This fields allows $(SolutionDir) style macros)")]
        public string AdditionalCommandLine { set; get; } = "";

        [UIDescription(Label = "Enable Warnings", Tooltip = "If true, the clang parser will output the warnings found.")]
        public bool EnableWarnings { set; get; } = false;
    }

    public class SettingsManager
    {
        private static readonly Lazy<SettingsManager> lazy = new Lazy<SettingsManager>(() => new SettingsManager());
        public static SettingsManager Instance { get { return lazy.Value; } }

        const string SettingsName = "StructLayoutSettings.json";

        private SolutionEvents solutionEvents; //Super important if not copied the events get disposed and never triggered
        public SolutionSettings Settings { get; set; }
        private string Filename{ set; get; }
        private Common.FileWatcher Watcher { set; get; }  = new Common.FileWatcher();

        private IServiceProvider ServiceProvider { set; get; }

        public void Initialize(IServiceProvider serviceProvider)
        {
            ThreadHelper.ThrowIfNotOnUIThread();

            ServiceProvider = serviceProvider;

            DTE2 applicationObject = ServiceProvider.GetService(typeof(SDTE)) as DTE2;
            Assumes.Present(applicationObject);
            solutionEvents = applicationObject.Events.SolutionEvents;
            solutionEvents.AfterClosing += RefreshFilename;
            solutionEvents.Opened += RefreshFilename;

            Watcher.FileWatchedChanged += Load;

            RefreshFilename();
        }

        private void RefreshFilename()
        {
            ThreadHelper.ThrowIfNotOnUIThread();

            string newFilename = null;

            DTE2 applicationObject = ServiceProvider.GetService(typeof(SDTE)) as DTE2;
            if (applicationObject != null && applicationObject.Solution != null)
            {
                string solutionDirRaw = applicationObject.Solution.FullName;
                if (solutionDirRaw.Length > 0)
                {
                    newFilename = (Path.HasExtension(solutionDirRaw) ? Path.GetDirectoryName(solutionDirRaw) : solutionDirRaw) + '/' + SettingsName;
                }
            }

            SetFilename(newFilename);
        }

        private void SetFilename(string str)
        {
            if (Filename != str)
            {
                Watcher.Unwatch();
                Filename = str;
                Load();
                Watcher.Watch(Filename);
            }
        }

        private void Load()
        {
            if (Filename != null && File.Exists(Filename))
            {
                string jsonString = File.ReadAllText(Filename);
                Settings = JsonConvert.DeserializeObject<SolutionSettings>(jsonString);
            }
        }

        public void Save()
        {
            if (Filename != null && Settings != null)
            {
                string jsonString = JsonConvert.SerializeObject(Settings, Formatting.Indented);
                File.WriteAllText(Filename, jsonString);
            }
        }

        private static T CloneJson<T>(T source)
        {
            if (Object.ReferenceEquals(source, null))
            {
                return default(T);
            }

            var deserializeSettings = new JsonSerializerSettings { ObjectCreationHandling = ObjectCreationHandling.Replace };
            return JsonConvert.DeserializeObject<T>(JsonConvert.SerializeObject(source), deserializeSettings);
        }

        public void OpenSettingsWindow()
        {
            SettingsWindow optionsWindow = new SettingsWindow(CloneJson<SolutionSettings>(Settings));
            optionsWindow.ShowDialog();
        }
    }
}
